/*
 * ipsec_ext AES cipher stubs (with optimized lib from libaes.sourceforge.net)
 *
 * Author: JuanJo Ciarlante <jjo-ipsec@mendoza.gov.ar>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 */
#include <linux/config.h>
#include <linux/version.h>

/*	
 *	special case: ipsec core modular with this static algo inside:
 *	must avoid MODULE magic for this file
 */
#if CONFIG_IPSEC_MODULE && CONFIG_IPSEC_EXT_AES_OPT
#undef MODULE
#endif

#include <linux/module.h>
#include <linux/init.h>

#include <linux/kernel.h> /* printk() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/string.h>

/* Check if __exit is defined, if not null it */
#ifndef __exit
#define __exit
#endif

/*	Low freeswan header coupling	*/
#include "ipsec_ext.h"
#include "libaes-opt/aes.h"
#include "libaes-opt/aes_cbc.h"

#define AES_CONTEXT_T aes_ctx
static int debug=0;
MODULE_PARM(debug, "i");
static int test=0;
MODULE_PARM(test, "i");

#define ESP_AES			12	/* truely _constant_  :)  */

#define ESP_AES_KEY_SZ		16 	/* 128 bit secret key */
#define ESP_AES_CBC_BLK_LEN	16	/* AES-CBC block size */

static int _aes_set_key(__u8 * key_e, const __u8 * key, int keysize) {
	int ret;
	AES_CONTEXT_T *ctx=(AES_CONTEXT_T*)key_e;
	ret=AES_set_key(ctx, (unsigned char *)key, keysize)!=0? 0: -EINVAL;
	if (debug > 0)
		printk(KERN_DEBUG "klips_debug:_aes_set_key:"
				"ret=%d key_e=%p key=%p keysize=%d\n",
				ret, key_e, key, keysize);
	return ret;
}
static int _aes_cbc_encrypt(__u8 * key_e, __u8 * in, __u8 * out, int ilen, const __u8 * iv, int encrypt) {
	AES_CONTEXT_T *ctx=(AES_CONTEXT_T*)key_e;
	if (debug > 0)
		printk(KERN_DEBUG "klips_debug:_aes_cbc_encrypt:"
				"key_e=%p in=%p out=%p ilen=%d iv=%p encrypt=%d\n",
				key_e, in, out, ilen, iv, encrypt);
	return AES_cbc_encrypt(ctx, in, out, ilen, iv, encrypt);
}
static struct ipsec_ext_enc ipsec_ext_AES = {
	ixt_version:	IPSEC_EXT_VERSION,
	ixt_module:	THIS_MODULE,
	ixt_refcnt:	ATOMIC_INIT(0),
	ixt_ext_type:	IPSEC_EXT_TYPE_ENCRYPT,
	ixt_alg_id: 	ESP_AES,
	ixt_name: 	"aes",
	ixt_blocksize:	ESP_AES_CBC_BLK_LEN,
	ixt_keyminbits:	ESP_AES_KEY_SZ*8,
	ixt_keymaxbits:	ESP_AES_KEY_SZ*8,
	ixt_e_keylen:	ESP_AES_KEY_SZ,
	ixt_e_ctx_size:	sizeof(AES_CONTEXT_T),
	ixt_e_set_key:	_aes_set_key,
	ixt_e_cbc_encrypt:_aes_cbc_encrypt,
};
	
IPSEC_EXT_MODULE_INIT( ipsec_aes_init )
{
	int ret, test_ret;
	ret=register_ipsec_ext_enc(&ipsec_ext_AES);
	printk(__FUNCTION__ "(ext_type=%d alg_id=%d name=%s): ret=%d\n", 
			ipsec_ext_AES.ixt_ext_type, 
			ipsec_ext_AES.ixt_alg_id, 
			ipsec_ext_AES.ixt_name, 
			ret);
	if (ret==0 && test) {
		test_ret=ipsec_ext_test(
				ipsec_ext_AES.ixt_ext_type,
				ipsec_ext_AES.ixt_alg_id, 
				test);
		printk(__FUNCTION__ "(ext_type=%d alg_id=%d): test_ret=%d\n", 
				ipsec_ext_AES.ixt_ext_type, 
				ipsec_ext_AES.ixt_alg_id, 
				test_ret);
	}
	return ret;
}
IPSEC_EXT_MODULE_EXIT( ipsec_aes_fini )
{
	unregister_ipsec_ext_enc(&ipsec_ext_AES);
	return;
}
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif

EXPORT_NO_SYMBOLS;
