/*
 * ipsec_ext RIPEMD hash stubs
 *
 * Author: JuanJo Ciarlante <jjo-ipsec@mendoza.gov.ar>
 * 
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.  See <http://www.fsf.org/copyleft/gpl.txt>.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 */
#include <linux/config.h>
#include <linux/version.h>

/*	
 *	special case: ipsec core modular with this static algo inside:
 *	must avoid MODULE magic for this file
 */
#if CONFIG_IPSEC_MODULE && CONFIG_IPSEC_EXT_RIPEMD
#undef MODULE
#endif

#include <linux/module.h>
#include <linux/init.h>

#include <linux/kernel.h> /* printk() */
#include <linux/errno.h>  /* error codes */
#include <linux/types.h>  /* size_t */
#include <linux/string.h>

/* Check if __exit is defined, if not null it */
#ifndef __exit
#define __exit
#endif

/*	Low freeswan header coupling	*/
#include "ipsec_ext.h"
#include "libripemd/ripemd.h"
#include "libripemd/hmac_ripemd.h"

static int debug=0;
MODULE_PARM(debug, "i");
static int test=0;
MODULE_PARM(test, "i");

#define AH_RIPEMD	8

static int _ripemd_hmac_set_key(__u8 * key_a, const __u8 * key, int keylen) {
	ripemd_hmac_context *hctx=(ripemd_hmac_context*)(key_a);
	ripemd_hmac_set_key(hctx, key, keylen);
	if (debug > 0)
		printk(KERN_DEBUG "klips_debug: " __FUNCTION__ "(): "
				"key_a=%p key=%p keysize=%d\n",
				key_a, key, keylen);
	return 0;
}
static int _ripemd_hmac_hash(__u8 * key_a, const __u8 * dat, int len, __u8 * hash, int hashlen) {
	ripemd_hmac_context *hctx=(ripemd_hmac_context*)(key_a);
	if (debug > 0)
		printk(KERN_DEBUG "klips_debug: " __FUNCTION__ "(): "
				"key_a=%p dat=%p len=%d hash=%p hashlen=%d\n",
				key_a, dat, len, hash, hashlen);
	ripemd_hmac_hash(hctx, dat, len, hash, hashlen);
	return 0;
}
static struct ipsec_ext_auth ipsec_ext_RIPEMD = {
	ixt_version:	IPSEC_EXT_VERSION,
	ixt_module:	THIS_MODULE,
	ixt_refcnt:	ATOMIC_INIT(0),
	ixt_ext_type:	IPSEC_EXT_TYPE_AUTH,
	ixt_alg_id: 	AH_RIPEMD,
	ixt_name: 	"ripemd",
	ixt_blocksize:	RIPEMD160_BLOCKSIZE,
	ixt_keyminbits:	160,
	ixt_keymaxbits:	160,
	ixt_a_keylen:	160/8,
	ixt_a_ctx_size:	sizeof(ripemd_hmac_context),
	ixt_a_hmac_set_key:	_ripemd_hmac_set_key,
	ixt_a_hmac_hash:	_ripemd_hmac_hash,
};
IPSEC_EXT_MODULE_INIT( ipsec_ripemd_init )
{
	int ret, test_ret;
	ret=register_ipsec_ext_auth(&ipsec_ext_RIPEMD);
	printk(__FUNCTION__ "(ext_type=%d alg_id=%d name=%s): ret=%d\n", 
			ipsec_ext_RIPEMD.ixt_ext_type, 
			ipsec_ext_RIPEMD.ixt_alg_id, 
			ipsec_ext_RIPEMD.ixt_name, 
			ret);
	if (ret==0 && test) {
		test_ret=ipsec_ext_test(
				ipsec_ext_RIPEMD.ixt_ext_type,
				ipsec_ext_RIPEMD.ixt_alg_id, 
				test);
		printk(__FUNCTION__ "(ext_type=%d alg_id=%d): test_ret=%d\n", 
				ipsec_ext_RIPEMD.ixt_ext_type, 
				ipsec_ext_RIPEMD.ixt_alg_id, 
				test_ret);
	}
	return ret;
}
IPSEC_EXT_MODULE_EXIT( ipsec_ripemd_fini )
{
	unregister_ipsec_ext_auth(&ipsec_ext_RIPEMD);
	return;
}
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
